import openai
import platform

__all__ = ("Translator")

class Translator(object):
    """
    """
    def __init__(self, token: str, param_mapping: dict):
        openai.api_key = token
        self._params = param_mapping
        self.os_name = platform.system()

    def translate(self, query):
        """
        """
        if not query:
            return None
        prompt = f"Convert this text to a programmatic command in {self.os_name}: {query}"
        self._params['prompt'] = prompt
        response = openai.Completion.create(**self._params)
        if response.choices:
            return response.choices[0].text
        else: 
            return None


from prompt_toolkit.completion import Completer, Completion

class ActionsCompleter(Completer):
    def get_completions(self, document, complete_event):
        yield Completion('Execute', start_position=-1 * len(document.text))
        yield Completion('Copy', start_position=-1 * len(document.text))
        yield Completion('Continue', start_position=-1 * len(document.text))


from terma.config import PROMPT_STYLE, USER_INPUT_PROMPT
from terma.completer import ActionsCompleter
from prompt_toolkit.formatted_text import FormattedText
from prompt_toolkit import print_formatted_text
from prompt_toolkit.validation import Validator


class PromptInterface:
    """
    Interface used to interact with the user
    via the CLI. The object will be responsible for
    getting inputs from the user and displaying all outputs. 
    """
    def __init__(self, session):
        """
        Sets the session, also intialises the
        completer used to present the set of choices to the
        user. 

        Parameters
        ----------
        session: The prompt_toolkit session.
        """
        self.session = session
        self.completer = ActionsCompleter()
        self.choiceValidator = Validator.from_callable(
            self._choice_validator,
            error_message='Invalid Followup action',
            move_cursor_to_end=True)
        self.defaultValidator = Validator.from_callable(
            self._non_empty_string_validator,
            error_message='Non empty string not proper input',
            move_cursor_to_end=True)

    def _choice_validator(self, text):
        """
        Function to validate the choice
        for followup action selected/written
        by the user. 
        """
        choices = ['Execute', 'Copy', 'Continue']
        return text in choices

    def _non_empty_string_validator(self, text):
        """
        The default validator to be used.
        It should only alert when input is empty
        """
        if not bool(text):
            return False
        else:
            return not text.isspace()
    
    def get_query_from_user(self):
        """
        Get the initial query from the user.

        Returns
        -------
        A plain text string denoting what the user wants
        """
        text = self.session.prompt(USER_INPUT_PROMPT, style=PROMPT_STYLE, validator=self.defaultValidator)
        return text

    def print_to_prompt(self, text: str):
        """
        Print a given response to the terminal. 
        """
        response_text = FormattedText([('class:response', text),])
        print_formatted_text(response_text, style=PROMPT_STYLE)
    
    def get_user_choice(self, text: str):
        """
        Enable completion to True
        Generates an automatic dropdown list of suggestions
        Enable completion to False

        Parameters
        ----------
        text: The text to be displayed before showing the dropdown

        Returns
        -------
        choice (str): follow up action choice
        """
        response_text = FormattedText([('class:response', f'{text} '),])
        self.session.complete_while_typing=True
        selected_option = self.session.prompt(response_text, completer=self.completer,
                                              pre_run=self.session.default_buffer.start_completion,
                                              validator=self.choiceValidator)
        self.session.complete_while_typing=False
        return selected_option
    


import os
from prompt_toolkit.styles import Style

def _get_token():
    '''
    Function to retrieve the OpenAI
    token from either the .terma file
    or from the envrionment as a env variable.
    '''
    home_dir = os.path.expanduser("~")
    terma_config_file = '.terma'
    config_file_path = os.path.join(home_dir, terma_config_file)
    if os.path.exists(config_file_path):
        with open(config_file_path, 'r') as f:
            token = f.readline()
    else:
        token = os.environ["OPENAI_TOKEN"]
    return token.strip() if token else None

# Open ai token to access the openai endpoints
OPENAI_TOKEN = _get_token()

# OpenAI model parameters that will be used while
# querying the openai endpoint
OPENAI_PARAM_MAPPING = {
    "model": "text-davinci-003",
    "prompt": "",
    "temperature": 0.7,
    "max_tokens": 100,
    "top_p": 1,
    "frequency_penalty": 0.2,
    "presence_penalty": 0
}


PROMPT_STYLE = Style.from_dict({
    # User input (default text).
    '':          '#00ff88',
    # Prompt.
    'username': '#884444',
    'pound':    '#00aa11 bold',
    'response': '#ff0066 bold',
    'terma': '#33FFB2',
    'action_choice': 'hidden'
})

USER_INPUT_PROMPT = [
    ('class:username', os.getlogin()),
    ('class:pound',    ' \u276F\u276F '),
]

MOTD = "Starting terma session \npress Ctrl+D or type \"quit()\" to exit"


from prompt_toolkit import print_formatted_text, PromptSession
from terma.utils import TermaUtil
from terma.utils import EmptyResponseException, UnknownActionException
from terma.prompt import PromptInterface
from terma.config import PROMPT_STYLE, MOTD


def coordinate(terma_util, action, command):
    if action == 'Continue':
        return True
    elif action == 'Copy':
        terma_util.copy(command)
    elif action == 'Execute':
        terma_util.execute(command)
    else:
        raise UnknownActionException()

def interact_with_user(prompt_interface, terma_util):
    """
    """
    user_query = prompt_interface.get_query_from_user()
    if not user_query:
        return True
    if user_query == 'quit()' or user_query == 'exit()':
        raise EOFError
    response = terma_util.translate_query(user_query)
    if not response:
        raise EmptyResponseException()
    # prompt_interface.print_to_prompt(response)
    choice = prompt_interface.get_user_choice(response)
    try:
        coordinate(terma_util, choice, response)
    except UnknownActionException:
        print('Unknown follow up action asked of me, so doing nothing instead')


def main():
    """
    """
    print(MOTD)
    session = PromptSession(style=PROMPT_STYLE)
    prompt_interface = PromptInterface(session)
    terma_util = TermaUtil()
    while True:
        try:
            interact_with_user(prompt_interface, terma_util)
        except KeyboardInterrupt:
            print_formatted_text('\n KeyboardInterrupt')
        except EOFError:
            print_formatted_text('\n Exiting...')
            break
        except EmptyResponseException:
            print_formatted_text('\n Unable to translate query please rephrase what you want to do')
            continue


if __name__ == "__main__":
    main()


from terma.config import OPENAI_PARAM_MAPPING, OPENAI_TOKEN, PROMPT_STYLE, USER_INPUT_PROMPT
from terma.translator import Translator
import pyperclip
import subprocess


class TermaUtil:
    """
    Class to coordinate all the activities inside
    the terma prompt session. 
    """
    def __init__(self):
        self.translator = Translator(OPENAI_TOKEN, OPENAI_PARAM_MAPPING)
    
    def translate_query(self, query: str):
        """
        """
        translation = self.translator.translate(query)
        return translation.strip()

    def copy(self, text: str):
        """
        Function that copies what is passed to it
        to the clipboard. 
        """
        pyperclip.copy(text)
        print("Copied to clipboard")

    def execute(self, command: str):
        """
        Function that uses subprocess to execute what
        is passed to it. 
        """
        if "|" in command:
            output, error = _piped_execution(command)
        else:
            output, error = _individual_execution(command)
        if(output):
            print(output.decode('utf-8'))
        if(error):
            print(error.decode('utf-8'))

class EmptyResponseException(Exception):
    """
    """
    pass

class UnknownActionException(Exception):
    """
    """
    pass

def _individual_execution(command):
    """
    Function to trigger an exceution of an indvidual
    command. 
    eg: "ls", "python -i", "ps -ef"
    """
    process = subprocess.Popen(command.split(' '))
    process.wait()
    return process.communicate()

def _piped_execution(command):
    """
    Function to trigger an execution of a command
    which has multiple subcommands grouped together
    via pipes

    eg: "ls -l | grep abc", "cat abc.txt | awk -4" 
    """
    sub_commands = [x.strip() for x in command.split('|')]
    current_input = None
    current_process = None
    for x in sub_commands:
        current_process = subprocess.Popen(x.split(' '),
                                           stdin=current_input,
                                           stdout=subprocess.PIPE)
        current_process.wait()
        current_input = current_process.stdout
    return current_process.communicate()



from setuptools import setup

setup(
    name='terma',
    version='1.0.3',
    author='Arjun Shenoy A V',
    author_email='arjunshenoyav@gmail.com',
    description='Terminal Assistant',
    long_description=open("README.md").read(),
    packages=['terma'],
    install_requires=[
        "openai",
        "prompt-toolkit >= 3.0.37",
        "pyperclip >= 1.8.2",
    ],
    entry_points={
        'console_scripts': [
            'terma=terma.app:main',
        ],
    },
)
